from setuptools import setup, find_packages

setup(name='ligeor',
      version='1.0b',
      description='LIght curve GEOmetRy of eclipsing binaries',
      url='https://github.com/gecheline/ligeor',
      author='Angela Kochoska',
      author_email='a.kochoska@gmail.com',
      license='MIT License',
      packages=find_packages(),
      zip_safe=False)